<?php

namespace Slack\Blockkit;

use Illuminate\Support\Collection;
use Slack\BlockKit;
use Slack\Exceptions\SlackException;
use Slack\Blockkit\Blocks\{Button,Text,TextEmoji};

/**
 * Class Blockkit Block
 * Represents a list of Blockit blocks
 *
 * @package Slack\Blockkit
 */
class Blocks extends BlockKit
{
	/**
	 * Actions
	 *
	 * @param array $items
	 * @return Blocks
	 */
	private function actions(array $items): self
	{
		$this->_data->push(array_merge(['type'=>'actions'],$items));

		return $this;
	}

	/**
	 * Add actions block
	 *
	 * @param Collection $elements
	 * @return Blocks
	 */
	public function addActionElements(Collection $elements): self
	{
		$this->actions(['elements'=>$elements]);

		return $this;
	}

	/**
	 * Add context items
	 *
	 * @param Collection $items
	 * @return Blocks
	 */
	public function addContextElements(Collection $items): self
	{
		return $this->context(['elements'=>$items]);
	}

	/**
	 * Add a bock divider
	 *
	 * @returns Blocks
	 */
	public function addDivider(): self
	{
		$this->_data->push(['type'=>'divider']);

		return $this;
	}

	/**
	 * Add a block header
	 *
	 * @param string $text
	 * @return Blocks
	 */
	public function addHeader(string $text): self
	{
		$this->_data->push(['type'=>'header','text'=>TextEmoji::item($text,TRUE)]);

		return $this;
	}

	/**
	 * @param Collection $options
	 * @param string     $action_id
	 * @return Collection
	 * @todo To Check
	 */
	public function addOverflow(Collection $options,string $action_id): Collection
	{
		return collect([
			'type'=>'overflow',
			'options'=>$options,
			'action_id'=>$action_id,
		]);
	}

	/**
	 * Add a section with accessories
	 *
	 * @param Text   $text
	 * @param Button $accessory
	 * @return Blocks
	 */
	public function addSectionAccessoryButton(Text $text,Button $accessory): self
	{
		return $this->section([
			'text'=>$text,
			'accessory'=>$accessory,
		]);
	}

	/**
	 * @param Text        $label
	 * @param string      $action
	 * @param Collection  $options
	 * @param string|null $default
	 * @return Blocks
	 * @deprecated
	 * @todo Look at addSectionAccessory
	 */
	public function addSelect(Text $label,string $action,Collection $options,string $default=NULL): self
	{
		$this->_data->put('type','section');
		$this->_data->put('text',$label);

		// Accessories
		$x = collect();
		$x->put('action_id',$action);
		$x->put('type','static_select');
		$x->put('options',$options->map(function ($item) {
			if (is_array($item))
				$item = (object)$item;

			return [
				'text'=>[
					'type'=>'plain_text',
					'text'=>(string)$item->name,
				],
				'value'=>(string)($item->value ?: $item->id)
			];
		}));

		if ($default) {
			$choice = $options->filter(function($item) use ($default) {
				if (is_array($item))
					$item = (object)$item;

				return ($item->value == $default) ? $item : NULL;
			})->filter()->pop();

			if ($choice) {
				$x->put('initial_option',[
					'text'=>TextEmoji::item($choice['name']),
					'value'=>(string)$choice['value']
				]);
			}
		}

		$this->_data->put('accessory',$x);

		return $this;
	}

	/**
	 * Add a section with fields
	 * @param Collection $items
	 * @return Blocks
	 */
	public function addSectionFields(Collection $items): self
	{
		return $this->section(['fields'=>$items]);
	}

	/**
	 * Generates a multiselect that queries back to the server for values
	 *
	 * @param Text   $label
	 * @param string $action
	 * @return Blocks
	 * @todo To Change - and rename addSectionMultiSelectInput()
	 * @deprecated
	 */
	public function addMultiSelectInput(Text $label,string $action): self
	{
		$this->_data->put('type','section');
		$this->_data->put('text',$label);
		$this->_data->put('accessory',[
			'action_id'=>$action,
			'type'=>'multi_external_select',
		]);

		return $this;
	}

	/**
	 * Add a section with a multi select list
	 *
	 * @param Text            $label
	 * @param string          $action
	 * @param Collection      $options
	 * @param Collection|null $selected
	 * @param int|null        $maximum
	 * @return Blocks
	 * @throws \Exception
	 * @note Slack only allows 100 items
	 */
	public function addSectionMultiSelectStaticInput(Text $label,string $action,Collection $options,Collection $selected=NULL,int $maximum=NULL): self
	{
		if ($options->count() > 100)
			throw new SlackException('Selection list cannot have more than 100 items.');

		$x = collect();
		$x->put('action_id',$action);
		$x->put('type','multi_static_select');
		$x->put('options',$options->transform(function ($item) {
			return ['text'=>TextEmoji::item($item->name),'value'=>(string)$item->id];
		}));

		if ($selected and $selected->count())
			$x->put('initial_options',$selected->transform(function ($item) {
				return ['text'=>TextEmoji::item($item->name),'value'=>(string)$item->id];
			}));

		if ($maximum)
			$x->put('max_selected_items',$maximum);

		return $this->section([
			'text' => $label,
			'accessory' => $x
		]);
	}

	/**
	 * Add a section with just text
	 *
	 * @param Text $text
	 * @return Blocks
	 */
	public function addSectionText(Text $text): self
	{
		return $this->section(['text'=>$text]);
	}

	/**
	 * A context block
	 *
	 * @param array $items
	 * @return Blocks
	 */
	private function context(array $items): self
	{
		$this->_data->push(array_merge(['type'=>'context'],$items));

		return $this;
	}

	/**
	 * Render the input dialog
	 *
	 * @param string $label
	 * @param string $action
	 * @param int    $minlength
	 * @param string $placeholder
	 * @param bool   $multiline
	 * @param string $hint
	 * @param string $initial
	 * @return $this
	 * @throws \Exception
	 * @deprecated - to optimize
	 */
	protected function input(string $label,string $action,int $minlength,string $placeholder='',bool $multiline=FALSE,string $hint='',string $initial='')
	{
		$this->_data->put('type','input');
		$this->_data->put('element',[
			'type'=>'plain_text_input',
			'action_id'=>$action,
			'placeholder'=>$this->text($placeholder ?: ' ','plain_text'),
			'multiline'=>$multiline,
			'min_length'=>$minlength,
			'initial_value'=>$initial,
		]);
		$this->_data->put('label',[
			'type'=>'plain_text',
			'text'=>$label,
			'emoji'=>true,
		]);
		$this->_data->put('optional',$minlength ? FALSE : TRUE);

		if ($hint)
			$this->_data->put('hint',$this->text($hint,'plain_text'));

		return $this;
	}

	/**
	 * A section block
	 *
	 * @param array $items
	 * @return Blocks
	 */
	private function section(array $items): self
	{
		$this->_data->push(array_merge(['type'=>'section'],$items));

		return $this;
	}

	/**
	 * Generates a single-line input dialog
	 *
	 * @param string $label
	 * @param string $action
	 * @param string $placeholder
	 * @param int    $minlength
	 * @param string $hint
	 * @param string $initial
	 * @return Blocks
	 * @throws \Exception
	 * @deprecated - to optimize
	 */
	public function addSingleLineInput(string $label,string $action,string $placeholder='',int $minlength=5,string $hint='',string $initial=''): self
	{
		return $this->input($label,$action,$minlength,$placeholder,FALSE,$hint,$initial);
	}

	/**
	 * Generates a multi-line input dialog
	 *
	 * @param string $label
	 * @param string $action
	 * @param string $placeholder
	 * @param int    $minlength
	 * @param string $hint
	 * @param string $initial
	 * @return Blocks
	 * @throws \Exception
	 * @deprecated - to optimize
	 */
	public function addMultiLineInput(string $label,string $action,string $placeholder='',int $minlength=20,string $hint='',string $initial=''): self
	{
		return $this->input($label,$action,$minlength,$placeholder,TRUE,$hint,$initial);
	}
}
