<?php

namespace Slack\Client;

use GuzzleHttp;
use Psr\Http\Message\ResponseInterface;
use React\EventLoop\LoopInterface;
use React\Promise\Deferred;
use React\Promise\PromiseInterface;

abstract class API
{
	/**
	* The base URL for API requests.
	*/
	const BASE_URL = 'https://slack.com/api/';

	/**
	 * @var string The Slack API token string.
	 */
	protected $token;

	/**
	 * @var GuzzleHttp\ClientInterface A Guzzle HTTP client.
	 */
	protected $httpClient;

	/**
	 * @var LoopInterface An event loop instance.
	 */
	protected $loop;

	/**
	 * Creates a new API client instance.
	 *
	 * @param LoopInterface                   $loop
	 * @param GuzzleHttp\ClientInterface|null $httpClient A Guzzle client instance to
	 *                                                    send requests with.
	 */
	public function __construct(LoopInterface $loop,GuzzleHttp\ClientInterface $httpClient = null)
	{
		$this->loop = $loop;
		$this->httpClient = $httpClient ?: new GuzzleHttp\Client();
	}

	/**
	 * Sets the Slack API token to be used during method calls.
	 *
	 * @param string $token The API token string.
	 */
	public function setToken($token)
	{
		$this->token = $token;
	}

	/**
	 * Sends an API request.
	 *
	 * @param string $method     The API method to call.
	 * @param array  $args       An associative array of arguments to pass to the
	 *                           method call.
	 * @param bool   $multipart  Whether to send as a multipart request. Default to false
	 * @param bool $callDeferred Whether to call the API asynchronous or not.
	 *
	 * @return PromiseInterface A promise for an API response.
	 */
	public function apiCall(string $method,array $args=[],bool $multipart=FALSE,bool $callDeferred=TRUE): PromiseInterface
	{
		// create the request url
		$requestUrl = self::BASE_URL.$method;

		// set the api token
		$args['token'] = $this->token;

		// send a post request with all arguments
		$requestType = $multipart ? 'multipart' : 'form_params';
		$requestData = $multipart ? $this->convertToMultipartArray($args) : $args;

		$promise = $this->httpClient->postAsync($requestUrl,[
			//$requestType => $requestData,
			'headers'=>[
				'Content-Type'=>'application/json',
				'Authorization'=>'Bearer '.$args['token'],
			]
		]);

		// Add requests to the event loop to be handled at a later date.
		if ($callDeferred) {
			$this->loop->futureTick(function () use ($promise) {
				$promise->wait();
			});

		} else {
			$promise->wait();
		}

		// When the response has arrived, parse it and resolve. Note that our
		// promises aren't pretty; Guzzle promises are not compatible with React
		// promises, so the only Guzzle promises ever used die in here and it is
		// React from here on out.
		$deferred = new Deferred();
		$promise->then(function (ResponseInterface $response) use ($deferred) {
			// get the response as a json object
			$payload = Payload::fromJson((string) $response->getBody());

			// check if there was an error
			if (isset($payload['ok']) && $payload['ok'] === TRUE) {
				$deferred->resolve($payload);

			} else {
				// make a nice-looking error message and throw an exception
				$niceMessage = ucfirst(str_replace('_', ' ', $payload['error']));
				$deferred->reject(new ApiException($niceMessage));
			}
		});

		return $deferred->promise();
	}

	private function convertToMultipartArray(array $options): array
	{
		$convertedOptions = [];

		foreach ($options as $key => $value) {
			$convertedOptions[] = [
				'name' => $key,
				'contents' => $value,
			];
		}

		return $convertedOptions;
	}
}
