<?php

namespace Slack\Interactive;

use Illuminate\Support\Arr;
use Illuminate\Support\Collection;
use Slack\Models\Channel;

/**
 * This class handles when the users responds with a Block Action event
 *
 * EG:
 * [type] => block_actions
 * [user] => stdClass Object
 * (
 *   [id] => {SLACKUSER}
 *   [username] => {SLACKUSER}
 *   [name] => {SLACKUSER}
 *   [team_id] => {SLACKTEAM}
 * )
 * [api_app_id] => A4TCZ007N
 * [token] => {SLACKTOKEN}
 * [container] => stdClass Object
 * (
 *   [type] => view
 *   [view_id] => V018HRRS38R
 * )
 * [trigger_id] => 1346041864311.39333245939.7c6adb7bca538143098386f07effa532
 * [team] => stdClass Object
 * (
 *   [id] => {SLACKTEAM}
 *   [domain] => {SLACKDOMAIN}
 * )
 * [view] => stdClass Object
 * (
 *   ...
 * )
 * [actions] => Array
 * (
 *   [0] => stdClass Object
 *   (
 *     [action_id] => faq_product
 *     [block_id] => IRFcN
 *     [text] => stdClass Object
 *     (
 *       [type] => plain_text
 *       [text] => ASK QUESTION
 *       [emoji] => 1
 *     )
 *     [value] => question_new
 *     [type] => button
 *     [action_ts] => 1600065294.860855
 *   )
 * )
 */
class BlockActions extends Base
{
	private const LOGKEY = 'IBA';

	public function __get($key)
	{
		switch ($key) {
			case 'actions':
				return object_get($this->_data,$key);

			case 'action_id':
				return object_get(Arr::get(object_get($this->_data,'actions'),$this->index),$key);

			// An event can have more than 1 action, each action can have 1 value.
			case 'action_key':
				return $this->action('action');

			case 'action_value':
				return $this->action('value');

			case 'callback_id':
				return object_get($this->_data,'view.callback_id');

			case 'keys':
				return collect(object_get($this->_data,'view.blocks'))->pluck('accessory.action_id');

			// For Block Actions that are messages
			case 'message_ts':
				return object_get($this->_data,'message.ts');

			case 'channel_id':
				return object_get($this->_data,'channel.id') ?: Channel::findOrFail($this->action('value'))->channel_id;

			case 'team_id':                                 // view.team_id represent workspace publishing view
				return object_get($this->_data,'user.team_id');

			case 'view_id':
				return object_get($this->_data,'view.id');

			case 'value':
				switch (Arr::get(object_get($this->_data,'actions'),$this->index)->type) {
					case 'external_select':
					case 'overflow':
					case 'static_select':
						return object_get(Arr::get(object_get($this->_data,'actions'),$this->index),'selected_option.value');
					case 'multi_static_select':
						return object_get(Arr::get(object_get($this->_data,'actions'),$this->index),'selected_options.value');
					default:
						return object_get(Arr::get(object_get($this->_data,'actions'),$this->index),$key);
				}

			case 'values':
				switch (Arr::get(object_get($this->_data,'actions'),$this->index)->type) {
					// @todo To Check
					case 'external_select':
					// @todo To Check
					case 'overflow':
					// @todo To Check
					case 'static_select':
						return count(object_get(Arr::get(object_get($this->_data,'actions'),$this->index),'selected_option'));
					case 'multi_static_select':
						return collect(object_get(Arr::get(object_get($this->_data,'actions'),$this->index),'selected_options'))->pluck('value');
					default:
						return count(object_get(Arr::get(object_get($this->_data,'actions'),$this->index),'value'));
				}

			case 'value_count':
				return count($this->values);

			default:
				return parent::__get($key);
		}
	}

	/**
	 * Separate out an action command to the id that the command relates to
	 *
	 * @param string $key
	 * @return string|null
	 */
	private function action(string $key): ?string
	{
		$regex = '/^([a-z_]+)\|([0-9]+)$/';
		$action = NULL;
		$value = NULL;

		// We only take the action up to the pipe symbol
		$action_key = object_get(Arr::get(object_get($this->_data,'actions'),$this->index),'action_id');

		if (preg_match($regex,$action_key)) {
			$action = preg_replace($regex,'$1',$action_key);
			$value = preg_replace($regex,'$2',$action_key);
		}

		switch ($key) {
			case 'action':
				return $action ?: $action_key;
			case 'value':
				return $value;
		}

		return NULL;
	}

	/**
	 * Some block actions are triggered by messages, and thus dont have a callback_id
	 *
	 * @return bool
	 */
	public function isMessage(): bool
	{
		return object_get($this->_data,'message') ? TRUE : FALSE;
	}
}
